-------------------------------------------------------------------------------
--! @file       ICEPOLE_Control.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Control unit for ICEPOLE
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity ICEPOLE_Control is
    generic (
        G_LBS_BYTES      : integer := 7
    );
    port (
        clk              : in  std_logic;
        rst              : in  std_logic;

        --! Input
        key_ready        : out std_logic;
        key_valid        : in  std_logic;
        key_update       : in  std_logic;
        decrypt          : in  std_logic;
        bdi_ready        : out std_logic;
        bdi_valid        : in  std_logic;        
        bdi_type         : in  std_logic_vector(3               -1 downto 0);
        bdi_eot          : in  std_logic;
        bdi_eoi          : in  std_logic;
        bdi_size         : in  std_logic_vector(G_LBS_BYTES+1   -1 downto 0);

        --! Datapath
        sel_state        : out std_logic_vector(2               -1 downto 0);
        en_state         : out std_logic;
        en_key           : out std_logic;
        round            : out std_logic_vector(4               -1 downto 0);
        is_nsec          : out std_logic;
        is_frame         : out std_logic;
        is_full          : out std_logic;
        sel_tag          : out std_logic;

        --! Output
        msg_auth_done    : out std_logic;
        bdo_ready        : in  std_logic;
        bdo_valid        : out std_logic
    );
end entity ICEPOLE_Control;

architecture behavior of ICEPOLE_Control is
    constant TOT_ROUND_INIT : integer := 12;
    constant TOT_ROUND_PROC : integer := 6;

    type state_type is (S_INIT, S_UPDATE_KEY, S_PROCESS, S_DONE, S_STALL);
	signal state			: state_type;
	signal nstate			: state_type;

    signal bdi_ad           : std_logic;
    signal en_round         : std_logic;
    signal clr_round        : std_logic;
    signal round_r          : std_logic_vector(4                -1 downto 0);
    signal rndcmp           : std_logic_vector(4                -1 downto 0);
	signal en_rndcmp		: std_logic;
    signal is_eoi           : std_logic;

    signal clr_ignore_key   : std_logic;
    signal set_ignore_key   : std_logic;
    signal ignore_key       : std_logic;

    --! Internal signals for output
    signal is_init          : std_logic;
    signal clr_isinit       : std_logic;
    signal clr_islast       : std_logic;
    signal is_decrypt       : std_logic;

    signal sel_state_s      : std_logic_vector(  2             -1 downto 0);
    signal sel_tag_s        : std_logic;
    signal en_state_s       : std_logic;    
    signal bdi_ready_s      : std_logic;
    signal bdo_valid_s      : std_logic;
    signal msg_auth_done_s  : std_logic;
begin

    round <= round_r;
    p_reg:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                state           <= S_INIT;
                round_r         <= (others => '0');
                rndcmp          <= (others => '0');
                is_frame        <= '0';
                is_full         <= '0';
                is_nsec         <= '0';
                ignore_key      <= '0';
            else
                state           <= nstate;

                --! Rounds
                if    (clr_round = '1') then
                    round_r     <= (others => '0');
                elsif (en_round  = '1') then
                    round_r     <= std_logic_vector(unsigned(round_r) + 1);
                end if;
                if  (clr_islast = '1') then
                    rndcmp   <= std_logic_vector(to_unsigned(TOT_ROUND_INIT-1,4));
                elsif (en_rndcmp = '1') then
                    rndcmp   <= std_logic_vector(to_unsigned(TOT_ROUND_PROC-1,4));
                end if;

                --! Flags
                if    (clr_islast = '1') then
                    is_init <= '1';
                elsif (clr_isinit = '1') then
                    is_init <= '0';
                end if;
                if (state = S_INIT) then
                    is_decrypt <= decrypt;
                end if;
                if (state = S_INIT) then
                    is_eoi <= '0';
                elsif (bdi_ready_s = '1' 
                       and bdi_type(2 downto 1) = BDI_TYPE_DAT
                       and (bdi_size(G_LBS_BYTES) /= '1' or bdi_eoi = '1')) 
                then                    
                    is_eoi <= '1';
                end if;
                if (clr_ignore_key = '1') then
                    ignore_key <= '0';
                elsif (set_ignore_key = '1') then
                    ignore_key <= '1';
                end if;

                --! Output registers
                if (bdi_ready_s = '1') then
                    if (bdi_size(G_LBS_BYTES) = '1') then
                        is_full <= '1';
                        if ((bdi_ad = '1' and bdi_eot = '1')
                            or (bdi_ad = '0' and bdi_eot = '0'))
                        then
                            is_frame <= '1';
                        else
                            is_frame <= '0';
                        end if;
                    else
                        is_full  <= '0';
                        is_frame <= '0';
                    end if;
                end if;
                is_nsec          <= clr_isinit;
                sel_state        <= sel_state_s;
                en_state         <= en_state_s;
                msg_auth_done    <= msg_auth_done_s;
                sel_tag          <= sel_tag_s;

                bdi_ready        <= bdi_ready_s;
                bdo_valid        <= bdo_valid_s;
            end if;
        end if;
    end process;
    bdi_ad <= '1' when bdi_type(2 downto 1) = BDI_TYPE_ASS else '0';

    p_state:
    process( state, bdi_valid, bdi_size, bdo_ready, key_valid,
             is_eoi, is_decrypt, is_init, ignore_key,
             round_r, rndcmp, bdi_ad, key_update)
    begin
        --! Default values
        nstate          <= state;
        en_state_s      <= '0';
        sel_state_s     <= "00";
        clr_round       <= '0';
        en_round        <= '0';
        bdi_ready_s     <= '0';
        bdo_valid_s     <= '0';
        en_rndcmp       <= '0';
        en_key          <= '0';
        key_ready       <= '0';
        clr_isinit      <= '0';
        clr_islast      <= '0';
        msg_auth_done_s <= '0';
        set_ignore_key  <= '0';
        clr_ignore_key  <= '0';
        sel_tag_s       <= '0';


        case state is
            when S_INIT =>
                clr_round      <= '1';
                clr_islast     <= '1';
                if ((key_update = '1' and ignore_key = '0')
                    or bdi_valid = '1')
                then
                    if (key_update = '1') then
                        nstate <= S_UPDATE_KEY;
                    elsif (bdi_valid = '1') then
                        en_state_s      <= '1';
                        sel_state_s     <= "01";
                        nstate          <= S_PROCESS;
                        bdi_ready_s     <= '1';
                    end if;
                end if;

            when S_UPDATE_KEY =>
                key_ready <= '1';
                if (key_valid = '1') then
                    en_key   <= '1';
                    en_round <= '1';
                    if (unsigned(round_r) = 3) then
                        set_ignore_key <= '1';
                        nstate <= S_INIT;
                    end if;
                end if;

            when S_PROCESS =>
                en_round    <= '1';
                en_state_s  <= '1';
                if (round_r = rndcmp) then
                    clr_round   <= '1';
                    if (is_init = '1') then
                        en_rndcmp   <= '1';
                        clr_isinit  <= '1';
                    else
                        nstate <= S_DONE;
                    end if;
                end if;

            when S_DONE =>
                if (bdi_valid = '1' and bdi_ad = '1') then
                    sel_state_s <= "10";
                    bdi_ready_s  <= '1';
                    en_state_s  <= '1';
                    nstate      <= S_PROCESS;
                elsif ((bdi_valid = '1' and bdi_ad = '0' 
                            and (bdo_ready = '1' or unsigned(bdi_size) = 0))
                        or (is_eoi = '1'
                            and ((is_decrypt = '0' and bdo_ready = '1')
                                or (is_decrypt ='1' and bdi_valid = '1'))))
                then
                    if (is_decrypt = '1') then
                        sel_state_s <= "11";
                    else
                        sel_state_s <= "10";
                    end if;
                    if (is_eoi = '1') then
                        clr_ignore_key <= '1';
                        sel_tag_s <= '1';
                        if (is_decrypt = '1') then
                            bdi_ready_s     <= '1';
                            msg_auth_done_s <= '1';
                            nstate          <= S_STALL;
                        else
                            bdo_valid_s     <= '1';
                            nstate          <= S_STALL;
                        end if;
                    else
                        bdi_ready_s  <= '1';
                        if (unsigned(bdi_size) /= 0) then
                            bdo_valid_s <= '1';
                        end if;
                        en_state_s  <= '1';
                        nstate      <= S_PROCESS;
                    end if;
                end if;

            when S_STALL =>
                nstate <= S_INIT;

        end case;
    end process;
end behavior;

