-------------------------------------------------------------------------------
--! @file       ICEPOLE_Datapath.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Datapath unit for ICEPOLE
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity ICEPOLE_Datapath is
    generic (
        G_BITRATE       : integer := 1024;
        G_NPUB_SIZE     : integer := 96
    );
    port    (
        clk             : in  std_logic;
        rst             : in  std_logic;

        --! Input
        bdi             : in  std_logic_vector(1024         -1 downto 0);
        key             : in  std_logic_vector(32           -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_BITRATE/8  -1 downto 0);

        --! Control
        sel_state       : in  std_logic_vector(2            -1 downto 0);
        en_state        : in  std_logic;
        en_key          : in  std_logic;
        round           : in  std_logic_vector(4            -1 downto 0);
        is_nsec         : in  std_logic;
        is_frame        : in  std_logic;
        is_full         : in  std_logic;
        sel_tag         : in  std_logic;

        --! Output
        msg_auth_valid : out std_logic;
        bdo            : out std_logic_vector(1024          -1 downto 0)
    );
end entity ICEPOLE_Datapath;

architecture structure of ICEPOLE_Datapath is
    constant    ZEROS   : std_logic_vector(1280             -1 downto 0)
                        := (others => '0');
    type     rc_type is array (0 to 15) of std_logic_vector(55 downto 0);
    constant rc_rom     : rc_type := (
             x"91A2B3C4D5E6F7", x"48D159E26AF37B",
             x"2468ACF13579BD", x"123456F89ABCDE",
             x"091A2BFC4D5E6F", x"048D15FE26AF37",
             x"02468AFF13579B", x"0123457F89ABCD",
             x"0091A2BFC4D5E6", x"0048D1DFE26AF3",
             x"002468EFF13579", x"001234F7F89ABC",
             x"00000000000000", x"00000000000000",
             x"00000000000000", x"00000000000000"
           );
    constant INIT       : std_logic_vector(1280             -1 downto 0) :=
        x"FF97A42D7F8E6FD4" & x"90FEE5A0A44647C4" & x"8C5BDA0CD6192E76"
            & x"AD30A6F71B19059C" & x"30935AB7D08FFC64" &
        x"EB5AA93F2317D635" & x"A9A6E6260D712103" & x"81A57C16DBCF555F"
            & x"43B831CD0347C826" & x"01F22F1A11A5569F" &
        x"05E5635A21D9AE61" & x"64BEFEF28CC970F2" & x"613670957BC46611"
            & x"B87C5A554FD00ECB" & x"8C3EE88A1CCF32C8" &
        x"940C7922AE3A2614" & x"1841F924A2C509E4" & x"16F53526E70465C2"
            & x"75F644E97F30A13B" & x"EAF1FF7B5CECA249";

	function switch_endian (
        signal      x       : std_logic_vector;
        constant    size    : integer;
        constant    w       : integer
    ) return std_logic_vector is
        constant xlow   : integer := x'low;
		constant xseg   : integer := size/w;
		constant wseg   : integer := w/8;
		variable retval : std_logic_vector(size-1 downto 0);
	begin
		for i in xseg downto 1 loop
			for j in 0 to wseg-1 loop
				retval(i*w-j*8+xlow-1 downto i*w-j*8+xlow-8)
                    := x(i*w-(wseg-1-j)*8+xlow-1
                         downto i*w-(wseg-1-j)*8+xlow-8);
			end loop;
		end loop;
		return(retval);
	end switch_endian;

    -- ! Signals declaration


    signal key_r              : std_logic_vector( 128       -1 downto 0);
    signal rc                 : std_logic_vector(  64       -1 downto 0);
    signal init_iv_key        : std_logic_vector(1280       -1 downto 0);
    signal rin                : std_logic_vector(1280       -1 downto 0);
    signal rout               : std_logic_vector(1280       -1 downto 0);

    signal bdi_lendian        : std_logic_vector(1024       -1 downto 0);
    signal bdo_lendian        : std_logic_vector(1024       -1 downto 0);
    signal bdo_s              : std_logic_vector(1024       -1 downto 0);
    signal key_lendian        : std_logic_vector( 128       -1 downto 0);
    signal npub_lendian       : std_logic_vector( 128       -1 downto 0);
    signal tag_lendian        : std_logic_vector( 128       -1 downto 0);
    signal npub_ext           : std_logic_vector( 128       -1 downto 0);
    signal tag_data           : std_logic_vector( 128       -1 downto 0);

    --! Debugging
    type state_type is array (natural range<>,natural range<>)
        of std_logic_vector(63 downto 0);

    signal istateState        : state_type(0 to 3, 0 to 4);
    signal ostateState        : state_type(0 to 3, 0 to 4);
    signal initState          : state_type(0 to 3, 0 to 4);
    signal routState          : state_type(0 to 3, 0 to 4);
    signal decState           : state_type(0 to 3, 0 to 4);
    signal encState           : state_type(0 to 3, 0 to 4);

    signal bdiState           : state_type(0 to 3, 0 to 3);
begin

    key_lendian <= switch_endian(key_r, 128, 64);
    G_NPUB96 : if G_NPUB_SIZE = 96  generate
        npub_ext <= bdi(G_BITRATE-1 downto G_BITRATE-G_NPUB_SIZE)
                    &  x"00000000";
    end generate;
    G_NPUB128: if G_NPUB_SIZE = 128 generate
        npub_ext <= bdi(G_BITRATE-1 downto G_BITRATE-G_NPUB_SIZE);
    end generate;
    npub_lendian  <= switch_endian(npub_ext, 128, 64);
    bdi_lendian   <= switch_endian(bdi, 1024, 64);

    init_iv_key <= (INIT(1279 downto 1216) xor key_lendian(127 downto 64)) &
                    INIT(1215 downto  960) &
                   (INIT( 959 downto  896) xor key_lendian( 63 downto  0)) &
                    INIT( 895 downto  640) &
                   (INIT( 639 downto  576) xor npub_lendian(127 downto 64)) &
                    INIT( 575 downto  320) &
                   (INIT( 319 downto  256) xor npub_lendian( 63 downto  0)) &
                    INIT( 255 downto    0);

    u_rc:
    process( clk )
    begin
        if rising_edge( clk ) then
            rc <= x"00" & rc_rom(to_integer(unsigned(round)));
            if (en_state = '1') then
                ostateState <= istateState;
            end if;
            if (en_key = '1') then
                key_r <= key_r(95 downto 0) & key;
            end if;
        end if;
    end process;

    u_round:
        entity work.ICEPOLE_Round(structure)
        port map (rin => rin, rc => rc, rout => rout);

    --! Note: The state has the following structure
    --!       S[0][0] ... S[0][4] S[1][0] .. S[3][4]
    --!
    --!       The value of bdi_xor and bdi_lendian are arranged as followed
    --!       S[0][0] .. S[3][0] S[0][1] .. S[3][3]
    --!       instead of
    --!       S[0][0] .. S[0][4] S[1][0] .. S[3][1]

    --! Map to state
    gRow: for i in 0 to 3 generate
        gCol: for j in 0 to 4 generate
            --! Column 0 to 3
            gMapX: if (j < 4) generate
                --! Encrypt/Decrypt operation
                encState(i,j) <= bdiState(i,j) xor ostateState(i,j);
                gBytes: for k in 0 to 7 generate
                    decState(i,j)(k*8+7 downto k*8)
                        <=  bdiState(i,j)(k*8+7 downto k*8)
                                when (bdi_valid_bytes(G_BITRATE/8-1-(k+j*32+i*8)) = '1')
                                else encState(i,j)(k*8+7 downto k*8);
                end generate;

                --! State mapping
                initState(i,j) <= init_iv_key(1280-i*320-j*64-1 downto 1280-i*320-j*64-64);
                g0: if (i = 0 and j = 0) generate
                    --! Note: First block deals with padding for NSEC
                    routState(i,j)(0)           <= rout(1280-i*320-j*64-64);
                    routState(i,j)(1)           <= rout(1280-i*320-j*64-63) xor is_nsec;
                    routState(i,j)(63 downto 2) <= rout(1280-i*320-j*64-1 downto 1280-i*320-j*64-62);
                end generate;
                gX: if (i /= 0 or j /= 0) generate
                    routState(i,j) <=        rout(1280-i*320-j*64-1 downto 1280-i*320-j*64-64);
                end generate;

                bdiState(i,j)  <= bdi_lendian(1024-i*64-j*256-1 downto 1024-i*64-j*256-64);

                --! State inverse mapping
                bdo_lendian(1024-i*64-j*256-1 downto 1024-i*64-j*256-64) <= encState(i,j);
                rin(1280-i*320-j*64-1 downto 1280-i*320-j*64-64)         <= ostateState(i,j);
            end generate;
            --! Last column
            gMap4: if (j = 4) generate
                gI0: if (i = 0) generate
                    --! Full block and frame bit
                    encState(i,j)(0)            <= ostateState(i,j)(0) xor is_frame;
                    encState(i,j)(1)            <= ostateState(i,j)(1) xor is_full;
                    encState(i,j)(63 downto 2)  <= ostateState(i,j)(63 downto 2);

                    decState(i,j)(0)            <= ostateState(i,j)(0) xor is_frame;
                    decState(i,j)(1)            <= ostateState(i,j)(1) xor is_full;
                    decState(i,j)(63 downto 2)  <= ostateState(i,j)(63 downto 2);
                end generate;
                gIX: if (i /= 0) generate
                    encState(i,j)  <= ostateState(i,j);
                    decState(i,j)  <= ostateState(i,j);
                end generate;

                --! State mapping
                initState(i,j) <= init_iv_key(1280-i*320-j*64-1 downto 1280-i*320-j*64-64);
                routState(i,j) <=        rout(1280-i*320-j*64-1 downto 1280-i*320-j*64-64);

                --! State inverse mapping
                rin(1280-i*320-j*64-1 downto 1280-i*320-j*64-64) <= ostateState(i,j);
            end generate;
        end generate;
    end generate;

    with sel_state(1 downto 0) select
    istateState <=  routState     when "00",
                    initState     when "01",
                    encState      when "10",
                    decState      when others;

    --! Output
    tag_lendian    <= ostateState(0,0) & ostateState(1,0);
    tag_data       <= switch_endian(tag_lendian, 128, 64);
    bdo_s          <= switch_endian(bdo_lendian, 1024, 64);
    bdo(1024-1 downto 1024-128) <= tag_data when sel_tag = '1' 
                                            else bdo_s(1024-1 downto 1024-128);
    bdo(1024-128-1 downto 0)    <= bdo_s(1024-128-1 downto 0);
    msg_auth_valid <= '1' when tag_data = bdi(G_BITRATE-1 downto G_BITRATE-128)
                          else '0';
end architecture structure;