-------------------------------------------------------------------------------
--! @file       ICEPOLE_Round.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Round unit for ICEPOLE
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity ICEPOLE_Round is
generic (
    STATE_SIZE          : integer := 1280);
port (
    rin     			: in std_logic_vector(STATE_SIZE-1 downto 0);
    rc			   	 	: in std_logic_vector(63 downto 0);
    rout    			: out std_logic_vector(STATE_SIZE-1 downto 0));
end ICEPOLE_Round;

architecture structure of ICEPOLE_Round is
    type t_sheet_matrix is array (0 to 3) of std_logic_vector(63 downto 0);
    type t_plane_matrix is array (0 to 4) of std_logic_vector(63 downto 0);
    type t_state_matrix is array (0 to 3) of t_plane_matrix;

    signal ss  : t_state_matrix;
    signal mu  : t_state_matrix;
    signal rp  : t_state_matrix;
    signal psi : t_state_matrix;
    signal temp1, temp2 : t_sheet_matrix;

    type t_rotate is array (0 to 3, 0 to 4) of integer range 0 to 63;
    constant ROTATE : t_rotate := ((0, 36, 3, 41, 18), (1, 44, 10, 45, 2), (62, 6, 43, 15, 61), (28, 55, 25, 21, 56));
begin
    -- # Convert to matrix ss(y)(x) based on http://keccak.noekeon.org/Keccak-f-PiecesOfState.png
	gMatrixPlane:
	for x in 0 to 3 generate
		gMatrixSheet:
		for y in 0 to 4 generate
			ss(x)(y) <= rin(STATE_SIZE-(5*x+y)*64-1 downto STATE_SIZE-(5*x+y)*64-64);
		end generate;
	end generate;

    -- # Mu step
    -- # E
    mu(0)(4) <= ss(0)(3) xor ss(1)(4) xor ss(2)(4) xor ss(3)(4);
    mu(0)(3) <= ss(0)(2) xor ss(1)(3) xor ss(2)(3) xor ss(3)(3);
    mu(0)(2) <= ss(0)(4) xor ss(0)(1) xor ss(1)(2) xor ss(2)(2) xor ss(3)(2);
    mu(0)(1) <= ss(0)(0) xor ss(1)(1) xor ss(2)(1) xor ss(3)(1);
    mu(0)(0) <= ss(0)(4) xor ss(1)(0) xor ss(2)(0) xor ss(3)(0);
    -- # F
    mu(1)(4) <= ss(0)(4) xor ss(1)(4) xor ss(2)(0) xor ss(3)(3);
    mu(1)(3) <= ss(0)(3) xor ss(1)(3) xor ss(2)(4) xor ss(3)(2);
    mu(1)(2) <= ss(0)(2) xor ss(1)(2) xor ss(2)(3) xor ss(3)(4) xor ss(3)(1);
    mu(1)(1) <= ss(0)(1) xor ss(1)(1) xor ss(2)(2) xor ss(2)(0) xor ss(3)(0);
    mu(1)(0) <= ss(0)(0) xor ss(1)(0) xor ss(2)(1) xor ss(3)(4);
    -- # G
    mu(2)(4) <= ss(0)(4) xor ss(1)(3) xor ss(2)(4) xor ss(3)(0);
    mu(2)(3) <= ss(0)(3) xor ss(1)(2) xor ss(2)(3) xor ss(3)(4);
    mu(2)(2) <= ss(0)(2) xor ss(1)(4) xor ss(1)(1) xor ss(2)(2) xor ss(3)(3);
    mu(2)(1) <= ss(0)(1) xor ss(1)(0) xor ss(2)(1) xor ss(3)(2) xor ss(3)(0);
    mu(2)(0) <= ss(0)(0) xor ss(1)(4) xor ss(2)(0) xor ss(3)(1);
    -- # H
    mu(3)(4) <= ss(0)(4) xor ss(1)(0) xor ss(2)(3) xor ss(3)(4);
    mu(3)(3) <= ss(0)(3) xor ss(1)(4) xor ss(2)(2) xor ss(3)(3);
    mu(3)(2) <= ss(0)(2) xor ss(1)(3) xor ss(2)(4) xor ss(2)(1) xor ss(3)(2);
    mu(3)(1) <= ss(0)(1) xor ss(1)(2) xor ss(1)(0) xor ss(2)(0) xor ss(3)(1);
    mu(3)(0) <= ss(0)(0) xor ss(1)(1) xor ss(2)(4) xor ss(3)(0);

    -- # Rho and Pi step
    gPi:
    for x in 0 to 3 generate
        gPiInner:
        for y in 0 to 4 generate
            rp((x+y) mod 4)((((x+y) mod 4)+y+1) mod 5) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(x)(y)), ROTATE(x,y)));
        end generate;
    end generate;

    -- # Psi & Kappa (Iota)
    gPsi:
    for x in 0 to 3 generate
        temp1(x)      <=     rp(x)(0) and     rp(x)(1) and     rp(x)(2) and     rp(x)(3) and     rp(x)(4);
        temp2(x)      <= not rp(x)(0) and not rp(x)(1) and not rp(x)(2) and not rp(x)(3) and not rp(x)(4);
        gKappa:
        if x  = 0 generate
            psi(x)(0) <= rp(x)(0) xor (not rp(x)(1) and rp(x)(2)) xor temp1(x) xor temp2(x) xor rc;
        end generate;
        gNotKappa:
        if x /= 0 generate
            psi(x)(0) <= rp(x)(0) xor (not rp(x)(1) and rp(x)(2)) xor temp1(x) xor temp2(x);
        end generate;
        psi(x)(1)     <= rp(x)(1) xor (not rp(x)(2) and rp(x)(3)) xor temp1(x) xor temp2(x);
        psi(x)(2)     <= rp(x)(2) xor (not rp(x)(3) and rp(x)(4)) xor temp1(x) xor temp2(x);
        psi(x)(3)     <= rp(x)(3) xor (not rp(x)(4) and rp(x)(0)) xor temp1(x) xor temp2(x);
        psi(x)(4)     <= rp(x)(4) xor (not rp(x)(0) and rp(x)(1)) xor temp1(x) xor temp2(x);
    end generate;

	gInvMatrixPlane:
	for x in 0 to 3 generate
		gInvMatrixSheet:
		for y in 0 to 4 generate
			rout(STATE_SIZE-(5*x+y)*64-1 downto STATE_SIZE-(5*x+y)*64-64) <= psi(x)(y);
		end generate;
	end generate;
end structure;


architecture structure_v2 of ICEPOLE_Round is
    type t_sheet_matrix is array (0 to 3) of std_logic_vector(63 downto 0);
    type t_plane_matrix is array (0 to 4) of std_logic_vector(63 downto 0);
    type t_state_matrix is array (0 to 3) of t_plane_matrix;

    signal ss  : t_state_matrix;
    signal mu  : t_state_matrix;
    signal rp  : t_state_matrix;
    signal psi : t_state_matrix;
begin
    -- # Convert to matrix ss(y)(x) based on http://keccak.noekeon.org/Keccak-f-PiecesOfState.png
	gMatrixPlane:
	for x in 0 to 3 generate
		gMatrixSheet:
		for y in 0 to 4 generate
			ss(x)(y) <= rin(STATE_SIZE-(5*x+y)*64-1 downto STATE_SIZE-(5*x+y)*64-64);
		end generate;
	end generate;

    -- # Mu step
    -- # E
    mu(0)(4) <= ss(0)(3) xor ss(1)(4) xor ss(2)(4) xor ss(3)(4);
    mu(0)(3) <= ss(0)(2) xor ss(1)(3) xor ss(2)(3) xor ss(3)(3);
    mu(0)(2) <= ss(0)(4) xor ss(0)(1) xor ss(1)(2) xor ss(2)(2) xor ss(3)(2);
    mu(0)(1) <= ss(0)(0) xor ss(1)(1) xor ss(2)(1) xor ss(3)(1);
    mu(0)(0) <= ss(0)(4) xor ss(1)(0) xor ss(2)(0) xor ss(3)(0);
    -- # F
    mu(1)(4) <= ss(0)(4) xor ss(1)(4) xor ss(2)(0) xor ss(3)(3);
    mu(1)(3) <= ss(0)(3) xor ss(1)(3) xor ss(2)(4) xor ss(3)(2);
    mu(1)(2) <= ss(0)(2) xor ss(1)(2) xor ss(2)(3) xor ss(3)(4) xor ss(3)(1);
    mu(1)(1) <= ss(0)(1) xor ss(1)(1) xor ss(2)(2) xor ss(2)(0) xor ss(3)(0);
    mu(1)(0) <= ss(0)(0) xor ss(1)(0) xor ss(2)(1) xor ss(3)(4);
    -- # G
    mu(2)(4) <= ss(0)(4) xor ss(1)(3) xor ss(2)(4) xor ss(3)(0);
    mu(2)(3) <= ss(0)(3) xor ss(1)(2) xor ss(2)(3) xor ss(3)(4);
    mu(2)(2) <= ss(0)(2) xor ss(1)(4) xor ss(1)(1) xor ss(2)(2) xor ss(3)(3);
    mu(2)(1) <= ss(0)(1) xor ss(1)(0) xor ss(2)(1) xor ss(3)(2) xor ss(3)(0);
    mu(2)(0) <= ss(0)(0) xor ss(1)(4) xor ss(2)(0) xor ss(3)(1);
    -- # H
    mu(3)(4) <= ss(0)(4) xor ss(1)(0) xor ss(2)(3) xor ss(3)(4);
    mu(3)(3) <= ss(0)(3) xor ss(1)(4) xor ss(2)(2) xor ss(3)(3);
    mu(3)(2) <= ss(0)(2) xor ss(1)(3) xor ss(2)(4) xor ss(2)(1) xor ss(3)(2);
    mu(3)(1) <= ss(0)(1) xor ss(1)(2) xor ss(1)(0) xor ss(2)(0) xor ss(3)(1);
    mu(3)(0) <= ss(0)(0) xor ss(1)(1) xor ss(2)(4) xor ss(3)(0);

    -- # Rho and Pi step
    rp(0)(0) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(0)(4)), 18));
    rp(0)(1) <=                                       mu(0)(0);    
    rp(0)(2) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(3)(1)), 55));
    rp(0)(3) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(2)(2)), 43));
    rp(0)(4) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(1)(3)), 45));
                
    rp(1)(0) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(2)(3)), 15));
    rp(1)(1) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(1)(4)), 2 ));
    rp(1)(2) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(1)(0)), 1 ));
    rp(1)(3) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(0)(1)), 36));
    rp(1)(4) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(3)(2)), 25));
                
    rp(2)(0) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(0)(2)), 3 ));
    rp(2)(1) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(3)(3)), 21));
    rp(2)(2) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(2)(4)), 61));
    rp(2)(3) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(2)(0)), 62));    
    rp(2)(4) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(1)(1)), 44));
                
    rp(3)(0) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(2)(1)), 6 ));
    rp(3)(1) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(1)(2)), 10));
    rp(3)(2) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(0)(3)), 41));
    rp(3)(4) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(3)(0)), 28));
    rp(3)(3) <= std_logic_vector(ROTATE_LEFT(unsigned(mu(3)(4)), 56));

    -- # Psi & Kappa (Iota)
    psi(0)(0) <= rp(0)(0) xor (not rp(0)(1) and rp(0)(2)) xor (rp(0)(0) and rp(0)(1) and rp(0)(2) and rp(0)(3) and rp(0)(4)) xor (not rp(0)(0) and not rp(0)(1) and not rp(0)(2) and not rp(0)(3) and not rp(0)(4)) xor rc;
    psi(0)(1) <= rp(0)(1) xor (not rp(0)(2) and rp(0)(3)) xor (rp(0)(0) and rp(0)(1) and rp(0)(2) and rp(0)(3) and rp(0)(4)) xor (not rp(0)(0) and not rp(0)(1) and not rp(0)(2) and not rp(0)(3) and not rp(0)(4));
    psi(0)(2) <= rp(0)(2) xor (not rp(0)(3) and rp(0)(4)) xor (rp(0)(0) and rp(0)(1) and rp(0)(2) and rp(0)(3) and rp(0)(4)) xor (not rp(0)(0) and not rp(0)(1) and not rp(0)(2) and not rp(0)(3) and not rp(0)(4));
    psi(0)(3) <= rp(0)(3) xor (not rp(0)(4) and rp(0)(0)) xor (rp(0)(0) and rp(0)(1) and rp(0)(2) and rp(0)(3) and rp(0)(4)) xor (not rp(0)(0) and not rp(0)(1) and not rp(0)(2) and not rp(0)(3) and not rp(0)(4));
    psi(0)(4) <= rp(0)(4) xor (not rp(0)(0) and rp(0)(1)) xor (rp(0)(0) and rp(0)(1) and rp(0)(2) and rp(0)(3) and rp(0)(4)) xor (not rp(0)(0) and not rp(0)(1) and not rp(0)(2) and not rp(0)(3) and not rp(0)(4));
    psi(1)(0) <= rp(1)(0) xor (not rp(1)(1) and rp(1)(2)) xor (rp(1)(0) and rp(1)(1) and rp(1)(2) and rp(1)(3) and rp(1)(4)) xor (not rp(1)(0) and not rp(1)(1) and not rp(1)(2) and not rp(1)(3) and not rp(1)(4));
    psi(1)(1) <= rp(1)(1) xor (not rp(1)(2) and rp(1)(3)) xor (rp(1)(0) and rp(1)(1) and rp(1)(2) and rp(1)(3) and rp(1)(4)) xor (not rp(1)(0) and not rp(1)(1) and not rp(1)(2) and not rp(1)(3) and not rp(1)(4));
    psi(1)(2) <= rp(1)(2) xor (not rp(1)(3) and rp(1)(4)) xor (rp(1)(0) and rp(1)(1) and rp(1)(2) and rp(1)(3) and rp(1)(4)) xor (not rp(1)(0) and not rp(1)(1) and not rp(1)(2) and not rp(1)(3) and not rp(1)(4));
    psi(1)(3) <= rp(1)(3) xor (not rp(1)(4) and rp(1)(0)) xor (rp(1)(0) and rp(1)(1) and rp(1)(2) and rp(1)(3) and rp(1)(4)) xor (not rp(1)(0) and not rp(1)(1) and not rp(1)(2) and not rp(1)(3) and not rp(1)(4));
    psi(1)(4) <= rp(1)(4) xor (not rp(1)(0) and rp(1)(1)) xor (rp(1)(0) and rp(1)(1) and rp(1)(2) and rp(1)(3) and rp(1)(4)) xor (not rp(1)(0) and not rp(1)(1) and not rp(1)(2) and not rp(1)(3) and not rp(1)(4));
    psi(2)(0) <= rp(2)(0) xor (not rp(2)(1) and rp(2)(2)) xor (rp(2)(0) and rp(2)(1) and rp(2)(2) and rp(2)(3) and rp(2)(4)) xor (not rp(2)(0) and not rp(2)(1) and not rp(2)(2) and not rp(2)(3) and not rp(2)(4));
    psi(2)(1) <= rp(2)(1) xor (not rp(2)(2) and rp(2)(3)) xor (rp(2)(0) and rp(2)(1) and rp(2)(2) and rp(2)(3) and rp(2)(4)) xor (not rp(2)(0) and not rp(2)(1) and not rp(2)(2) and not rp(2)(3) and not rp(2)(4));
    psi(2)(2) <= rp(2)(2) xor (not rp(2)(3) and rp(2)(4)) xor (rp(2)(0) and rp(2)(1) and rp(2)(2) and rp(2)(3) and rp(2)(4)) xor (not rp(2)(0) and not rp(2)(1) and not rp(2)(2) and not rp(2)(3) and not rp(2)(4));
    psi(2)(3) <= rp(2)(3) xor (not rp(2)(4) and rp(2)(0)) xor (rp(2)(0) and rp(2)(1) and rp(2)(2) and rp(2)(3) and rp(2)(4)) xor (not rp(2)(0) and not rp(2)(1) and not rp(2)(2) and not rp(2)(3) and not rp(2)(4));
    psi(2)(4) <= rp(2)(4) xor (not rp(2)(0) and rp(2)(1)) xor (rp(2)(0) and rp(2)(1) and rp(2)(2) and rp(2)(3) and rp(2)(4)) xor (not rp(2)(0) and not rp(2)(1) and not rp(2)(2) and not rp(2)(3) and not rp(2)(4));
    psi(3)(0) <= rp(3)(0) xor (not rp(3)(1) and rp(3)(2)) xor (rp(3)(0) and rp(3)(1) and rp(3)(2) and rp(3)(3) and rp(3)(4)) xor (not rp(3)(0) and not rp(3)(1) and not rp(3)(2) and not rp(3)(3) and not rp(3)(4));
    psi(3)(1) <= rp(3)(1) xor (not rp(3)(2) and rp(3)(3)) xor (rp(3)(0) and rp(3)(1) and rp(3)(2) and rp(3)(3) and rp(3)(4)) xor (not rp(3)(0) and not rp(3)(1) and not rp(3)(2) and not rp(3)(3) and not rp(3)(4));
    psi(3)(2) <= rp(3)(2) xor (not rp(3)(3) and rp(3)(4)) xor (rp(3)(0) and rp(3)(1) and rp(3)(2) and rp(3)(3) and rp(3)(4)) xor (not rp(3)(0) and not rp(3)(1) and not rp(3)(2) and not rp(3)(3) and not rp(3)(4));
    psi(3)(3) <= rp(3)(3) xor (not rp(3)(4) and rp(3)(0)) xor (rp(3)(0) and rp(3)(1) and rp(3)(2) and rp(3)(3) and rp(3)(4)) xor (not rp(3)(0) and not rp(3)(1) and not rp(3)(2) and not rp(3)(3) and not rp(3)(4));
    psi(3)(4) <= rp(3)(4) xor (not rp(3)(0) and rp(3)(1)) xor (rp(3)(0) and rp(3)(1) and rp(3)(2) and rp(3)(3) and rp(3)(4)) xor (not rp(3)(0) and not rp(3)(1) and not rp(3)(2) and not rp(3)(3) and not rp(3)(4));

	gInvMatrixPlane:
	for x in 0 to 3 generate
		gInvMatrixSheet:
		for y in 0 to 4 generate
			rout(STATE_SIZE-(5*x+y)*64-1 downto STATE_SIZE-(5*x+y)*64-64) <= psi(x)(y);
		end generate;
	end generate;
end structure_v2;

